@@ -1,11 +1,12 @@
-import re
+from itertools import product
+from textsearch import TextSearch
 
 
 contractions_dict = {
     "ain't": "are not",
     "aren't": "are not",
-    "can't": "cannot",
-    "can't've": "cannot have",
+    "can't": "can not",
+    "can't've": "can not have",
     "'cause": "because",
     "could've": "could have",
     "couldn't": "could not",
@@ -23,6 +24,7 @@ contractions_dict = {
     "he'll've": "he will have",
     "he's": "he is",
     "how'd": "how did",
+    "how're": "how are",
     "how'd'y": "how do you",
     "how'll": "how will",
     "how's": "how is",
@@ -119,14 +121,23 @@ contractions_dict = {
     "you'll've": "you shall have",
     "you're": "you are",
     "you've": "you have",
-    "doin'": "doing",
-    "goin'": "going",
-    "nothin'": "nothing",
-    "somethin'": "somethin'"
 }
 
+for month in [
+    "january",
+    "february",
+    "march",
+    "april",
+    "june",
+    "july",
+    "august",
+    "september",
+    "october",
+    "november",
+    "december",
+]:
+    contractions_dict[month[:3] + "."] = month
 
-contractions_re_keys = [x.replace("'", "['’]") for x in contractions_dict]
 contractions_dict.update({k.replace("'", "’"): v for k, v in contractions_dict.items()})
 
 
@@ -137,66 +148,97 @@ leftovers_dict = {
     "'d": " would",
     "'ll": " will",
     "'re": " are",
+    "'em": " them",
+    "doin'": "doing",
+    "goin'": "going",
+    "nothin'": "nothing",
+    "somethin'": "something",
+    "havin'": "having",
+    "doin'": "doing",
+    "lovin'": "loving",
+    "'coz": "because",
+    "thats": "that is",
+    "whats": "what is",
 }
 
 leftovers_dict.update({k.replace("'", "’"): v for k, v in leftovers_dict.items()})
 
-safety_keys = set(["he's", "he'll", "we'll", "we'd", "it's", "i'd", "we'd"])
+safety_keys = set(["he's", "he'll", "we'll", "we'd", "it's", "i'd", "we'd", "we're", "i'll"])
 
-unsafe_dict = {k.replace("'", ""): v for k, v in contractions_dict.items()
-               if k.lower() not in safety_keys}
 
-slang = {
+def get_combinations(tokens, joiners):
+    combs = []
+    combs.append(tokens)
+    results = []
+    for option in combs:
+        option = [[x] for x in option]
+        option = intersperse(option, joiners)
+        for c in product(*option):
+            results.append("".join(c))
+    return results
+
+
+def intersperse(lst, item):
+    result = [item] * (len(lst) * 2 - 1)
+    result[0::2] = lst
+    return result
+
+
+# have to create all the possibilities of ' and nothing
+unsafe_dict = {}
+for k, v in contractions_dict.items():  # contractions_dict.items():
+    if k.lower() in safety_keys:
+        continue
+    if "'" not in k:
+        continue
+    tokens = k.split("'")
+    for comb in get_combinations(tokens, ["", "'"]):
+        unsafe_dict[comb] = v
+
+
+slang_dict = {
     "ima": "I am going to",
     "gonna": "going to",
     "gotta": "got to",
+    "wanna": "want to",
+    "woulda": "would have",
     "gimme": "give me",
     "asap": "as soon as possible",
     "u": "you",
-    "r ": "are "
+    "r ": "are ",
 }
 
-unsafe_dict.update(slang)
+slang_dict.update(unsafe_dict)
 
-leftovers_re = re.compile('|'.join(sorted(leftovers_dict.keys())), re.IGNORECASE)
-contractions_re = re.compile('|'.join(sorted(contractions_re_keys)), re.IGNORECASE)
-unsafe_re = re.compile(r"\b" + r"\b|\b".join(sorted(unsafe_dict)) + r"\b", re.IGNORECASE)
+ts_leftovers = TextSearch("ignore", "norm")
+ts_leftovers.add(contractions_dict)
+ts_leftovers.add(leftovers_dict)
 
+ts_leftovers_slang = TextSearch("ignore", "norm")
+ts_leftovers_slang.add(contractions_dict)
+ts_leftovers_slang.add(leftovers_dict)
+ts_leftovers_slang.add(slang_dict)
 
-def _replacer(dc):
-    def replace(match):
-        v = match.group()
-        if v in dc:
-            return dc[v]
-        v = v.lower()
-        if v in dc:
-            return dc[v]
-        return v
-    return replace
+ts_slang = TextSearch("ignore", "norm")
+ts_slang.add(contractions_dict)
+ts_slang.add(slang_dict)
 
-slang_re = re.compile(r"\b" + r"\b|\b".join(sorted(list(slang) +
-                                                   list(unsafe_dict))) + r"\b", re.IGNORECASE)
+ts_basic = TextSearch("ignore", "norm")
+ts_basic.add(contractions_dict)
 
-LIM_RE = re.compile("['’]")
-
-rc = _replacer(contractions_dict)
-rl = _replacer(leftovers_dict)
-ru = _replacer(unsafe_dict)
+replacers = {
+    (True, False): ts_leftovers,
+    (True, True): ts_leftovers_slang,
+    (False, True): ts_slang,
+    (False, False): ts_basic,
+}
 
 
 def fix(s, leftovers=True, slang=True):
-    # when not expecting a lot of matches, this will be 30x faster
-    # otherwise not noticeably slower even in benchmarks
-    if not LIM_RE.search(s):
-        if slang and slang_re.search(s):
-            pass
-        else:
-            # ensure str like expected from re.sub
-            return str(s)
-    s = contractions_re.sub(rc, s)
-    if leftovers:
-        s = leftovers_re.sub(rl, s)
-    if slang:
-        s = unsafe_re.sub(ru, s)
+    ts = replacers[(leftovers, slang)]
+    return ts.replace(s)
 
-    return s
+
+def add(key, value):
+    for ts in replacers.values():
+        ts.add(key, value)
